#include <amxmodx>
#include <fakemeta>
#include <engine>
#include <fun>
#include <hamsandwich>
#include <xs>
#include <orpheu>
#include <orpheu_advanced>

#if !defined _orpheu_included
	#assert "orpheu.inc library required ! Get it from https://forums.alliedmods.net/showthread.php?t=116393"
#endif

#if !defined _orpheu_advanced_included
	#assert "orpheu_advanced.inc library required ! Get it from https://forums.alliedmods.net/showthread.php?t=116393"
#endif

#pragma unused heGrenadeAmmoIndex

#define pluginName    "Grenade Manager"
#define pluginVersion "0.2"
#define pluginAuthor  "HamletEagle"

#define invalidHook OrpheuHook:0

#define MAX_ITEM_TYPES 6
#define MAX_AMMO_SLOTS 32

#if AMXX_VERSION_NUM < 183
	#define INT_BYTES        4 
	#define BYTE_BITS        8 

	stock bool:get_pdata_bool(ent, charbased_offset, intbase_linuxdiff = 5) 
	{ 
		return !!( get_pdata_int(ent, charbased_offset / INT_BYTES, intbase_linuxdiff) & (0xFF<<((charbased_offset % INT_BYTES) * BYTE_BITS)) ) 
	} 
#endif


enum _:Grenades
{
	flashBang,
	heGrenade,
	smokeGrenade
}

enum
{
	primaryWeaponSlot = 1,
	pistolSlot,
	knifeSlot,
	grenadeSlot,
	c4Slot
}

enum _:Functions
{
	OrpheuFunction:installGamerules,
	OrpheuFunction:createNamedEntity,
	OrpheuFunction:cvar_DirectSet,
	OrpheuFunction:packWeapon,
	OrpheuFunction:kill,
	OrpheuFunction:getNextBestWeapon,
	OrpheuFunction:ammoInventory,
	OrpheuFunction:dropPlayerItem
}

new const weaponBoxClassName[] = "weaponbox"

new const m_rgAmmo_CBasePlayer        [MAX_AMMO_SLOTS] = {376, 377, ...}
new const m_rgpPlayerItems_CBasePlayer[MAX_ITEM_TYPES] = {367, 368, ...}
new const m_rgpPlayerItems_CWeaponBox [MAX_ITEM_TYPES] = {34 , 35 , ...}

new const grenadeClassNames[Grenades][] =
{
	"weapon_flashbang",
	"weapon_hegrenade",
	"weapon_smokegrenade"
}

new const grenadeAmmoNames[Grenades][]=
{
	"Flashbang", 
	"HEGrenade",
	"SmokeGrenade"
}

new const grenadesBaseAmmo[Grenades] = {2, 1, 1}

const m_pActiveItem	     = 373 //active item
const m_bIsVip           = 837 //true if a player is VIP
const m_pNext       	 = 42  //next linked weapon 
const m_iId	       	     = 43  //weapon index 
const m_iPrimaryAmmoType = 49  //primary ammo index into players m_rgAmmo[]
const m_pfnThink    	 = 4   //think function for entity
const XoPlayer      	 = 5   //player    linux offset = windows offset + 5
const XoCBasePlayerItem  = 4   //item      linux offset = windows offset + 4
const XoCWeaponBox		 = 4   //weaponbox linux offset = windows offset + 4

new flashBangAmmoIndex    
new heGrenadeAmmoIndex    
new smokeGrenadeAmmoIndex

new cvarDropGrenades
new cvarDropGrenadesOnDeath
new cWeaponBoxKillFuncAddress
new weaponBoxClassNameOffset
new g_pGameRules

new gmsgTextMsg
new gmsgWeaponList

new grenadesDefaultAmmo    [Grenades]
new cvarMaxGrenades        [Grenades]
new cvarMaxGrenadesValues  [Grenades]

new OrpheuFunction:handleGameFunction[Functions]

new OrpheuHook:handleAmmoInventoryHook
new OrpheuHook:handledropPlayerItemHook
new HamHook:handleTouchHook
new HamHook:handleGiveAmmoHook

new handleGrenadeDropForward
new weaponListMessageIndex

public plugin_precache()
{
	gmsgWeaponList = get_user_msgid("WeaponList")
	weaponListMessageIndex = register_message(gmsgWeaponList, "OnWeaponList_Message")
	
	handleGameFunction[installGamerules] = OrpheuGetFunction("InstallGameRules")
	OrpheuRegisterHook(handleGameFunction[installGamerules], "OnInstallGameRules", OrpheuHookPost)
}

public OnInstallGameRules()
{
	g_pGameRules = OrpheuGetReturn()    
	handleGameFunction[getNextBestWeapon] = OrpheuGetFunctionFromObject(g_pGameRules, "GetNextBestWeapon", "CGameRules")
}

public plugin_cfg()
{
	register_plugin
	(
		.plugin_name = pluginName,
		.version     = pluginVersion,
		.author      = pluginAuthor
	)

	cvarMaxGrenades[flashBang   ] = register_cvar("max_flashbangs"   , "2")
	cvarMaxGrenades[heGrenade   ] = register_cvar("max_hegrenades"   , "1")
	cvarMaxGrenades[smokeGrenade] = register_cvar("max_smokegrenades", "1")
	cvarDropGrenades              = register_cvar("drop_grenades"    , "1") 
	cvarDropGrenadesOnDeath		  = register_cvar("drop_grenades_on_death", "1")
	
	//RegisterHam(Ham_Killed, "player", "CBasePlayer_Killed", false)
	register_event("DeathMsg","CBasePlayer_Killed","a") 
	
	new bool:alterAmmo, defaultAmmo
	for(new i; i < sizeof cvarMaxGrenades; i++)
	{
		defaultAmmo              = grenadesBaseAmmo[i]
		cvarMaxGrenadesValues[i] = get_pcvar_num(cvarMaxGrenades[i])

		if(cvarMaxGrenadesValues[i] != defaultAmmo)
		{
			alterAmmo = true
		}
		
		grenadesDefaultAmmo[i] = defaultAmmo
	}
	
	new const cBasePlayer[] = "CBasePlayer"
	new const cWeaponBox [] = "CWeaponBox"

	handleGameFunction[cvar_DirectSet   ] = OrpheuGetFunction("Cvar_DirectSet"             )
	handleGameFunction[createNamedEntity] = OrpheuGetFunction("CREATE_NAMED_ENTITY"        )
	handleGameFunction[ammoInventory    ] = OrpheuGetFunction("AmmoInventory" , cBasePlayer)
	handleGameFunction[dropPlayerItem   ] = OrpheuGetFunction("DropPlayerItem", cBasePlayer)
	handleGameFunction[kill             ] = OrpheuGetFunction("Kill"          , cWeaponBox )
	handleGameFunction[packWeapon       ] = OrpheuGetFunction("PackWeapon"    , cWeaponBox )
	
	OrpheuRegisterHook(handleGameFunction[cvar_DirectSet], "OnCvar_DirectSet" , OrpheuHookPre)
	
	handleGiveAmmoHook = RegisterHam(Ham_GiveAmmo, "player"          , "CBasePlayerPlayer_GiveAmmo", false)
	handleTouchHook    = RegisterHam(Ham_Touch   , weaponBoxClassName, "CWeaponBox_Touch"          , false)
	
	if(alterAmmo)
	{
		handleAmmoInventoryHook = OrpheuRegisterHook(handleGameFunction[ammoInventory], "CBasePlayer_AmmoInventory" , OrpheuHookPre)
	}
	else 
	{
		DisableHamForward(handleGiveAmmoHook)
	}
	
	if(get_pcvar_num(cvarDropGrenades))
	{
		handledropPlayerItemHook = OrpheuRegisterHook(handleGameFunction[dropPlayerItem], "CBasePlayer_DropPlayerItem", OrpheuHookPre)
	}
	else 
	{
		DisableHamForward(handleTouchHook)
	}
	
	/*
		| When a weaponbox entity is created, game sets think to CWeaponBox::Kill and pev_nextthink to gametime() + 300.0
		| This creates a timer that remove the box after 300 seconds
		| Kill() basically get all items from weaponbox and remove them + remove the box
		| Calling think on a custom weaponbox will do nothing, because game does not know what to do with it when it thinks
		| Here we retrieve CWeaponBox::Kill function address so we can set later m_pfnThink offset and let game handle the weaponbox.
		| Another way would be to create a bot, give him a weapon, force drop, hook Touch and retrieve the address.
	*/
	cWeaponBoxKillFuncAddress = OrpheuGetFunctionAddress(handleGameFunction[kill])

	#if AMXX_VERSION_NUM < 183
		weaponBoxClassNameOffset = engfunc(EngFunc_AllocString, weaponBoxClassName)
	#endif
	
	handleGrenadeDropForward = CreateMultiForward("OnGrenadeDrop", ET_STOP, FP_CELL, FP_STRING, FP_CELL, FP_CELL, FP_CELL)
	
	unregisterMessage()
}


public CBasePlayer_Killed()
{
	new victim = read_data(2)
	if(is_user_connected(victim))
	{
		if(get_pcvar_num(cvarDropGrenadesOnDeath))
		{
			new name[32]
			get_user_name(victim, name, 31)
			for(new i; i < sizeof grenadeClassNames; i++)
			{
				if(user_has_weapon(victim, get_weaponid(grenadeClassNames[i])))
				{
					server_print("%s drop: %s", name,  grenadeClassNames[i])
					CBasePlayer_DropPlayerItem(victim, grenadeClassNames[i])
				}
			}
		}
	}
}

public plugin_end()
{
	DestroyForward(handleGrenadeDropForward)
}

public OnWeaponList_Message()
{
	/*
		WeaponList 
		string	WeaponName
		byte	PrimaryAmmoID
		byte	PrimaryAmmoMaxAmount
		byte	SecondaryAmmoID
		byte	SecondaryAmmoMaxAmount
		byte	SlotID
		byte	NumberInSlot
		byte	WeaponID
		byte	Flags
	*/
	new const weaponIndexParam      = 8
	new const primaryAmmoIndexParam = 2
	
	new ammoIndex
	for(new i; i < sizeof grenadesDefaultAmmo; i++)
	{
		ammoIndex = get_msg_arg_int(primaryAmmoIndexParam)
		switch(get_msg_arg_int(weaponIndexParam))
		{
			case CSW_FLASHBANG:
			{
				flashBangAmmoIndex    = ammoIndex
			}
			case CSW_HEGRENADE:
			{
				heGrenadeAmmoIndex    = ammoIndex
			}
			case CSW_SMOKEGRENADE:
			{
				smokeGrenadeAmmoIndex = ammoIndex
			}
		}
	}
}

unregisterMessage()
{
	unregister_message(gmsgWeaponList, weaponListMessageIndex)
}

public OnCvar_DirectSet(const cvarPointer, const cvarValue[])
{
	new cvarNewValue = str_to_num(cvarValue)
	new cvarOldValue
	
	if(cvarPointer == cvarDropGrenades)
	{
		cvarOldValue = get_pcvar_num(cvarDropGrenades)
		if(cvarNewValue != cvarOldValue)
		{
			if(cvarNewValue)
			{
				handledropPlayerItemHook = OrpheuRegisterHook(handleGameFunction[dropPlayerItem], "CBasePlayer_DropPlayerItem", OrpheuHookPre)
				EnableHamForward(handleTouchHook)
			}
			else 
			{
				OrpheuUnregisterHook(handledropPlayerItemHook)
				DisableHamForward(handleTouchHook)
				handledropPlayerItemHook = invalidHook
			}
		}
	}
	else 
	{
		new bool:unHook = true
		
		for(new i, k; i < sizeof cvarMaxGrenades; i++)
		{
			if(cvarPointer == cvarMaxGrenades[i])
			{
				cvarOldValue = cvarMaxGrenadesValues[i]
				if(cvarNewValue != cvarOldValue)
				{
					cvarMaxGrenadesValues[i] = cvarNewValue
					
					for(k = 0; k < sizeof grenadesDefaultAmmo; k++)
					{
						if(cvarMaxGrenadesValues[k] != grenadesDefaultAmmo[k])
						{
							unHook = false
							break
						}
					}

					if(unHook)
					{
						OrpheuUnregisterHook(handleAmmoInventoryHook)
						DisableHamForward(handleGiveAmmoHook)
						handleAmmoInventoryHook = invalidHook
					}
					else 
					{
						if(!handleAmmoInventoryHook)
						{
							handleAmmoInventoryHook = OrpheuRegisterHook(handleGameFunction[ammoInventory], "CBasePlayer_AmmoInventory" , OrpheuHookPre)
						}
						
						EnableHamForward(handleGiveAmmoHook)
					}
				}
				
				break
			}
		}
	}
}

/*
	| Called in order to drop player item(also from "drop" command)
	| If itemName is null then player wants to drop active item.
	| Inside game there is a check with CanDrop() which simply return true/false depending on item 
	| We could hook this function and set it's return value to true, but game is not designed to drop grenades so such situation is not properly handled.
	| For example, if you have more grenades you won't drop just one, you will drop all of them inside one weaponbox
	| Considering default game behaviour, this is only an issue with flashbang, but given that this plugin also change max grenade ammout 
	| it was a must to redo partially the function.
*/
public OrpheuHookReturn:CBasePlayer_DropPlayerItem(const id, const itemName[])
{
	new weaponEntity = FM_NULLENT, activeItem = FM_NULLENT
	new weaponClassName[32]
	
	if(is_user_connected(id))
	{
		weaponEntity = get_pdata_cbase(id, m_rgpPlayerItems_CBasePlayer[grenadeSlot], XoPlayer)
		activeItem   = get_pdata_cbase(id, m_pActiveItem, XoPlayer)
			
		while(pev_valid(weaponEntity))
		{
			if(itemName[0] != EOS)
			{
				pev(weaponEntity, pev_classname, weaponClassName, charsmax(weaponClassName))
				if(equal(itemName, weaponClassName))
				{
					break
				}
			}
			else
			{
				if(weaponEntity == activeItem)
				{
					break
				}
			}
				
			weaponEntity = get_pdata_cbase(weaponEntity, m_pNext, XoCBasePlayerItem)
		}
			
		if(pev_valid(weaponEntity))
		{
			if(get_pdata_bool(id, m_bIsVip, XoPlayer))
			{
				if(!gmsgTextMsg)
				{
					gmsgTextMsg = get_user_msgid("TextMsg")
				}
					
				message_begin(MSG_ONE, gmsgTextMsg, .player = id)
				{
					write_byte(print_center)
					write_string("#Weapon_Cannot_Be_Dropped")
					message_end()
				}

				return OrpheuSupercede
			}
				
			new ammoIndex = ExecuteHam(Ham_Item_PrimaryAmmoIndex, weaponEntity)
			if(ammoIndex != -1)
			{
				new userBackpackAmmo = get_pdata_int(id, m_rgAmmo_CBasePlayer[ammoIndex], XoPlayer)
					
				if(userBackpackAmmo)
				{
					new weaponIndex = get_pdata_int(weaponEntity, m_iId, XoCBasePlayerItem)
					new dummyEntity = FM_NULLENT
					new weaponBoxEntity 
						
					if(weaponClassName[0] == EOS)
					{
						//If we are here then player wants to drop active item, get it's classname
						pev(weaponEntity, pev_classname, weaponClassName, charsmax(weaponClassName))
					}
						
					do 
					{
						/*
							| Create a new entity, this will be packed into the box instead of player grenade
							| If we pack player grenade, then item will be removed, even if he has more grenades in inventory
						*/
						dummyEntity = create_entity(weaponClassName)
					}
					while(!pev_valid(dummyEntity))
						
					set_pev(dummyEntity, pev_spawnflags, pev(dummyEntity, pev_spawnflags) | SF_NORESPAWN)
					set_pdata_int(dummyEntity, m_iPrimaryAmmoType, ammoIndex, XoCBasePlayerItem)
						
					ExecuteHam(Ham_Spawn, dummyEntity)
						
					/*
						| Due to optimizations entities created by game are cached and searching functions use the hash table to find them.
						| weaponbox entities should be removed when round ends, but if you use default amxx functions like create_entity() 
						| entity won't be inserted in hash table => game will ignore it and default behaviour will be breaked.
					*/
					#if AMXX_VERSION_NUM < 183
						weaponBoxEntity = OrpheuCallSuper(handleGameFunction[createNamedEntity], weaponBoxClassNameOffset)
					#else 
						weaponBoxEntity = cs_create_entity(weaponBoxClassName)
					#endif
						
					if(!pev_valid(weaponBoxEntity))
					{
						return OrpheuIgnored
					}
						
					new returnValue
					ExecuteForward(handleGrenadeDropForward, returnValue, id, weaponClassName, dummyEntity, weaponBoxEntity, userBackpackAmmo)
						
					if(returnValue >= PLUGIN_HANDLED)
					{
						remove_entity(weaponBoxEntity)
						remove_entity(dummyEntity)
						return OrpheuIgnored
					}
						
					new Float:playerOrigin[3]; pev(id, pev_origin, playerOrigin)
					new Float:playerAngles[3]; pev(id, pev_angles, playerAngles)
						
					new Float:forwardDirection[3]
					new Float:forwardDirection2[3]
					engfunc(EngFunc_MakeVectors, playerAngles)
					global_get(glb_v_forward, forwardDirection)
						
					xs_vec_mul_scalar(forwardDirection, 10.0, forwardDirection)
					xs_vec_add(playerOrigin, forwardDirection, playerOrigin)
					
					set_pev(weaponBoxEntity, pev_origin, playerOrigin)
					ExecuteHam(Ham_Spawn, weaponBoxEntity)
						
					playerAngles[0] = 0.0
					playerAngles[2] = 0.0
					set_pev(weaponBoxEntity, pev_angles, playerAngles)
						
					set_pdata_int(weaponBoxEntity, m_pfnThink, cWeaponBoxKillFuncAddress, 0)
					set_pev(weaponBoxEntity, pev_nextthink, get_gametime() + 300)
						
					set_pdata_int(id, m_rgAmmo_CBasePlayer[ammoIndex], userBackpackAmmo - 1, XoPlayer)
						
					if(userBackpackAmmo - 1 <= 0)
					{
						OrpheuCallSuper(handleGameFunction[getNextBestWeapon], g_pGameRules, id, weaponEntity)
							
						if(activeItem == weaponEntity)
						{
							/*
								| If player drop active item, then we can Holster() it and let game do some of the function calls for us.
								| Grenades have a special holster. When ammo is 0 it cleans weapon hud and calls RemovePlayerItem().
							*/
							ExecuteHamB(Ham_Item_Holster, weaponEntity, 0)
						}
						else 
						{
							/*
								| There is nothing to holster, we need to do above manually
							*/
							new userWeapons = pev(id, pev_weapons)
							userWeapons &= ~ (1 << weaponIndex)
							set_pev(id, pev_weapons, userWeapons)
							ExecuteHamB(Ham_RemovePlayerItem, id, weaponEntity)
						}
					}
						
					OrpheuCallSuper(handleGameFunction[packWeapon], weaponBoxEntity, dummyEntity)
						
					xs_vec_copy(forwardDirection, forwardDirection2)
					xs_vec_mul_scalar(forwardDirection , 30.0, forwardDirection)
					xs_vec_mul_scalar(forwardDirection2, 10.0, forwardDirection2)
					xs_vec_add(forwardDirection, forwardDirection2, forwardDirection)
					set_pev(weaponBoxEntity, pev_velocity, forwardDirection)
						
					new modelName[32]
					GetCSModelName(weaponIndex, modelName, charsmax(modelName))
					engfunc(EngFunc_SetModel, weaponBoxEntity, modelName)
				}
					
				return OrpheuSupercede
			}
		}
	}
	return OrpheuIgnored
}

/*
	| Called from CArmoury::ArmouryTouch() and BuyItem()
	| It is used to check if player can get another grenade.
	| Since default values are hardcoded(2/1/1) and redoing the function will be painful
	| The best option is to alter the return value of AmmoInventory
*/
public OrpheuHookReturn:CBasePlayer_AmmoInventory(const id, const ammoIndex)
{
	if(flashBangAmmoIndex <= ammoIndex <= smokeGrenadeAmmoIndex)
	{
		
		new grenadeType = ammoIndex - flashBangAmmoIndex
		new maxGrenadeAmmount = cvarMaxGrenadesValues[grenadeType]
		
		if(maxGrenadeAmmount != grenadesDefaultAmmo[grenadeType])
		{
			new grenadesCount = get_pdata_int(id, m_rgAmmo_CBasePlayer[ammoIndex], XoPlayer)
			new returnValue = -1
			
			if(grenadesCount < maxGrenadeAmmount)
			{	
				returnValue = 0
			}
			else 
			{
				if(maxGrenadeAmmount < grenadesDefaultAmmo[grenadeType])
				{
					/*
						| Above is handled the case when cvar values are bigger than default values 
						| Here we take care of lower values.
					*/
					returnValue = grenadesDefaultAmmo[grenadeType] + 1
				}
			}
			
			if(returnValue != -1)
			{
				OrpheuSetReturn(returnValue)
				return OrpheuOverride
			}
		}
		
	}
	return OrpheuIgnored
}

public CBasePlayerPlayer_GiveAmmo(const id, const amount, const ammoName[], const maxAmmo)
{
	new maxGrenadeAmmount
	for(new i; i < sizeof grenadeAmmoNames; i++)
	{
		if(equal(grenadeAmmoNames[i], ammoName))
		{	
			maxGrenadeAmmount = cvarMaxGrenadesValues[i]
			
			if(maxGrenadeAmmount != grenadesDefaultAmmo[i])
			{
				SetHamParamInteger(4, maxGrenadeAmmount)
			}
			
			break
		}
	}
}

/*
	| As stated before, game was not coded to allow grenade drop and it gets pretty bugged when we change that.
	| For example, if you pickup one grenade you will receive two, because a new grenade entity is spawned.
	| Also, max amount of grenade is hardcoded, so we handle manually grenade touch case.
*/

public CWeaponBox_Touch(const weaponBoxEntity, const id)
{	
	if(pev_valid(weaponBoxEntity))
	{
		if(!(pev(weaponBoxEntity, pev_flags) & FL_ONGROUND))
		{
			return HAM_IGNORED
		}
		
		if(!is_user_alive(id))
		{
			return HAM_IGNORED
		}
		
		new weaponEntity = get_pdata_cbase(weaponBoxEntity, m_rgpPlayerItems_CWeaponBox[grenadeSlot], XoCWeaponBox)
		new ammoIndex, playerGrenades, maxGrenades, grenadeName[32], weaponIndex
		
		if(pev_valid(weaponEntity))
		{
			ammoIndex = ExecuteHam(Ham_Item_PrimaryAmmoIndex, weaponEntity)
			if(ammoIndex != -1)
			{
				playerGrenades = get_pdata_int(id, m_rgAmmo_CBasePlayer[ammoIndex], XoPlayer)
				maxGrenades    = 0

				weaponIndex = get_pdata_int(weaponEntity, m_iId, XoCBasePlayerItem)
				switch(weaponIndex)
				{
					case CSW_HEGRENADE:
					{
						copy(grenadeName, charsmax(grenadeName), grenadeClassNames[heGrenade])
						maxGrenades = cvarMaxGrenadesValues[heGrenade]
					}
					case CSW_SMOKEGRENADE:
					{
						copy(grenadeName, charsmax(grenadeName), grenadeClassNames[smokeGrenade])
						maxGrenades = cvarMaxGrenadesValues[smokeGrenade]
					}
					case CSW_FLASHBANG:
					{
						copy(grenadeName, charsmax(grenadeName), grenadeClassNames[flashBang])
						maxGrenades = cvarMaxGrenadesValues[flashBang]
					}
				}
					
				if(playerGrenades < maxGrenades && grenadeName[0] != EOS)
				{
					give_item(id, grenadeName)
					call_think(weaponBoxEntity)
				}
			}
			
			return HAM_SUPERCEDE
		}
	}
	
	return HAM_IGNORED
}

/*
	| Retrieve model to set on the box.
	| In CS, SetModel() is fired two times on a weaponbox.
	| First time, for w_weaponbox.mdl, then it set the model according to the packed weapon.
*/
GetCSModelName(const itemId, modelName[], const Size)
{
	switch(itemId)
	{
		case CSW_HEGRENADE   :	copy(modelName, Size, "models/w_hegrenade.mdl"   )
		case CSW_SMOKEGRENADE:	copy(modelName, Size, "models/w_smokegrenade.mdl")
		case CSW_FLASHBANG   :	copy(modelName, Size, "models/w_flashbang.mdl"   )
	}
}
